"""Main module."""
from multiprocessing import Pool
from typing import Dict, List
from urllib import parse

import requests

from scrapemove.models import (
    CombinedDetails,
    Property,
    PropertyDetailsScreenData,
    ResultsScreenData,
)

_DEFAULT_THREADPOOL = 1
_VALID_DOMAIN = "www.rightmove.co.uk"


def _validate_url(url: str):
    """Basic validation of the user-supplied URL"""
    parsed = parse.urlsplit(url)

    valid_protocols = ["http", "https"]
    if parsed.scheme not in valid_protocols:
        raise ValueError(f"Invalid scheme {parsed.scheme} (must be {valid_protocols})")

    if parsed.netloc != _VALID_DOMAIN:
        raise ValueError(f"Invalid domain: {parsed.netloc} (must be {_VALID_DOMAIN}")

    valid_paths = [
        f"/{p}/find.html"
        for p in ["property-to-rent", "property-for-sale", "new-homes-for-sale"]
    ]
    if parsed.path not in valid_paths:
        raise ValueError(f"Invalid path: {parsed.path} (must be {valid_paths}")


def _request_results_page(url: str) -> bytes:
    # Request
    r = requests.get(url)
    status_code, content = r.status_code, r.content
    if status_code != 200:
        raise RuntimeError(f"Request for {url} failed with: [{status_code}]: {content}")
    # Parse the data
    return content


def _load_results_page(url: str) -> ResultsScreenData:
    raw_content = _request_results_page(url)
    return ResultsScreenData.from_page_content(raw_content)


def _load_details_page(url: str) -> PropertyDetailsScreenData:
    raw_content = _request_results_page(url)
    return PropertyDetailsScreenData.from_page_content(raw_content)


def _build_url(url: str, params: Dict[str, str]) -> str:
    url_parts = list(parse.urlparse(url))
    query = dict(parse.parse_qs(url_parts[4]))
    query.update(params)
    url_parts[4] = parse.urlencode(query)
    return parse.urlunparse(url_parts)


def _flat_map(f, xs):
    ys = []
    for x in xs:
        ys.extend(f(x))
    return ys


def request(
    url: str, detailed=False, parallelism=_DEFAULT_THREADPOOL
) -> List[Property]:
    # Validate
    _validate_url(url)
    # Request the first page
    first_page = _load_results_page(url)
    # Extract the remaining pages
    next_page_urls = [
        f"{url}&index={option.value}" for option in first_page.pagination.options
    ]
    with Pool(parallelism) as p:
        subsequent_pages = p.map(_load_results_page, next_page_urls[1:])
    # Assemble together
    all_pages = [first_page] + subsequent_pages
    property_list = _flat_map(lambda r: r.properties, all_pages)
    if not detailed:
        return property_list
    # Request further details
    with Pool(parallelism) as p:
        details_pages = p.map(
            _load_details_page,
            [f"https://{_VALID_DOMAIN}{p.property_url}" for p in property_list],
        )
    details_list = [d.property_data for d in details_pages]
    merged_list = [
        CombinedDetails(property=p, additional_details=d)
        for p, d in zip(property_list, details_list)
    ]
    return merged_list



import json
from pathlib import Path

import pytest
from pydantic import ValidationError

from scrapemove.models import ResultsScreenData

_TEST_DIR = Path(__file__).parent
_TEST_DATA_DIR = _TEST_DIR / "test_data"
_RESULTS_SCREEN_DATA_FILE = _TEST_DATA_DIR / "results-screen-data.json"

with open(_RESULTS_SCREEN_DATA_FILE) as file:
    _RESULTS_SCREEN_DATA = json.load(file)


@pytest.mark.parametrize("data", [_RESULTS_SCREEN_DATA])
def test_parse_properties(data):
    try:
        ResultsScreenData(**data)
    except ValidationError as exc:
        assert False, f"Validation failed: {exc}"



from typing import List, Optional

from pydantic import HttpUrl

from scrapemove.models.common import (
    CamelCaseBaseModel,
    Location,
    Price,
    _parse_from_page,
)

class Sizing(CamelCaseBaseModel):
    unit: str
    display_unit: str
    minimum_size: float
    maximum_size: float


class Station(CamelCaseBaseModel):
    name: str
    types: List[str]
    distance: float
    unit: str


class PropertyDetails(CamelCaseBaseModel):
    title: str
    description: str
    share_description: str
    property_phrase: str
    key_features: List[str]
    postcode: str
    images: List[HttpUrl]
    floorplans: List[HttpUrl]
    nearest_stations: List[Station]
    brochures: List[str]
    sizings: List[Sizing]

    def __init__(self, **kwargs):
        """Remap some fields"""
        # Unnest product label
        kwargs["images"] = [i["url"] for i in kwargs["images"]]
        kwargs["floorplans"] = [i["url"] for i in kwargs["floorplans"]]
        kwargs["brochures"] = [i["url"] for i in kwargs["brochures"]]
        text = kwargs["text"]
        kwargs["title"] = text["pageTitle"]
        kwargs["description"] = text["description"]
        kwargs["shareDescription"] = text["shareDescription"]
        kwargs["propertyPhrase"] = text["propertyPhrase"]
        address = kwargs["address"]
        kwargs["postcode"] = f'{address.get("outcode")}{address.get("incode")}'
        super().__init__(**kwargs)


class PropertyDetailsScreenData(CamelCaseBaseModel):
    property_data: PropertyDetails

    @staticmethod
    def from_page_content(content: str) -> "PropertyDetailsScreenData":
        return _parse_from_page(
            content, r"window\.PAGE_MODEL =", PropertyDetailsScreenData
        )



import json
from pathlib import Path

import pytest
from pydantic import ValidationError

from scrapemove.models import PropertyDetailsScreenData

_TEST_DIR = Path(__file__).parent
_TEST_DATA_DIR = _TEST_DIR / "test_data"
_PROPERTY_DETAILS_SCREEN_DATA_FILE = (
    _TEST_DATA_DIR / "property-details-screen-data.json"
)

with open(_PROPERTY_DETAILS_SCREEN_DATA_FILE) as file:
    _PROPERTY_DETAILS_SCREEN_DATA = json.load(file)


@pytest.mark.parametrize("data", [_PROPERTY_DETAILS_SCREEN_DATA])
def test_parse_property_details(data):
    try:
        PropertyDetailsScreenData(**data)
    except ValidationError as exc:
        assert False, f"Validation failed: {exc}"



"""Data classes to enclose rightmove data"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field, HttpUrl

from scrapemove.models.common import (
    CamelCaseBaseModel,
    Location,
    Price,
    _parse_from_page,
)


class ListingUpdate(CamelCaseBaseModel):
    reason: str = Field(alias="listingUpdateReason")
    date: datetime = Field(alias="listingUpdateDate")


class PaginationOption(CamelCaseBaseModel):
    value: int
    description: int


class Pagination(CamelCaseBaseModel):
    first: int
    last: int
    options: List[PaginationOption]


class Property(CamelCaseBaseModel):
    id: int
    bedrooms: int
    bathrooms: Optional[int]
    number_of_images: int
    number_of_floorplans: int
    summary: Optional[str]
    display_address: Optional[str]
    country_code: Optional[str]
    location: Location
    property_sub_type: Optional[str]
    listing_update: ListingUpdate
    price: Price
    transaction_type: str
    product_label: str
    commercial: bool
    development: bool
    residential: bool
    students: bool
    auction: bool
    fees_apply: bool
    display_size: str
    property_url: str
    contact_url: str
    first_visible_date: datetime

    def __init__(self, **kwargs):
        """Remap some fields"""
        # Unnest product label
        kwargs["productLabel"] = kwargs["productLabel"]["productLabelText"]
        super().__init__(**kwargs)


class ResultsScreenData(CamelCaseBaseModel):
    properties: List[Property]
    pagination: Pagination
    resultCount: int

    @staticmethod
    def from_page_content(content: str) -> "ResultsScreenData":
        return _parse_from_page(content, r"window\.jsonModel =", ResultsScreenData)



from scrapemove.models.combined_details import CombinedDetails
from scrapemove.models.property_details_screen import (
    PropertyDetails,
    PropertyDetailsScreenData,
    Station,
)
from scrapemove.models.results_screen import (
    ListingUpdate,
    Location,
    Pagination,
    PaginationOption,
    Price,
    Property,
    ResultsScreenData,
)



import json
import re
from typing import Optional

from bs4 import BeautifulSoup
from inflection import camelize
from pydantic import BaseModel


class CamelCaseBaseModel(BaseModel):
    class Config:
        alias_generator = lambda s: camelize(s, False)


class Location(CamelCaseBaseModel):
    latitude: Optional[float]
    longitude: float


class Price(CamelCaseBaseModel):
    amount: Optional[int]
    currency_code: Optional[str]
    frequency: Optional[str]
    qualifier: Optional[str]


def _parse_from_page(content: str, regex, cls):
    soup = BeautifulSoup(content, "html.parser")
    script = soup.find("script", string=re.compile(regex))
    data_str = re.sub(regex, "", script.string)
    data_py = json.loads(data_str)
    return cls(**data_py)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scrapemove', 'scrapemove.models', 'tests']

package_data = \
{'': ['*'], 'tests': ['test_data/*']}

install_requires = \
['beautifulsoup4>=4.11.1,<5.0.0',
 'click',
 'inflection>=0.5.1,<0.6.0',
 'pydantic>=1.10.2,<2.0.0',
 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['scrapemove = scrapemove.cli:main']}

setup_kwargs = {
    'name': 'scrapemove',
    'version': '0.1.2',
    'description': 'Top-level package for scrapemove.',
    'long_description': '==========\nscrapemove\n==========\n\n\n.. image:: https://img.shields.io/pypi/v/scrapemove.svg\n        :target: https://pypi.python.org/pypi/scrapemove\n\n.. image:: https://img.shields.io/travis/briggySmalls/scrapemove.svg\n        :target: https://travis-ci.com/briggySmalls/scrapemove\n\n.. image:: https://readthedocs.org/projects/scrapemove/badge/?version=latest\n        :target: https://scrapemove.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nScraper for rightmove property data\n\n\n* Free software: MIT\n* Documentation: https://scrapemove.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Sam Briggs',
    'author_email': 'briggySmalls90@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/briggySmalls/scrapemove',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



"""Console script for scrapemove."""

import sys

import click

from scrapemove.scrapemove import request


@click.command()
@click.argument("url")
@click.option("--parallelism", type=int)
@click.option("--details/--no-details", default=False)
def main(url, parallelism, details):
    """Console script for scrapemove."""
    click.echo(request(url, detailed=details, parallelism=parallelism))
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from pydantic import BaseModel, HttpUrl

from scrapemove.models.property_details_screen import PropertyDetails
from scrapemove.models.results_screen import Property


class CombinedDetails(BaseModel):
    property: Property
    additional_details: PropertyDetails

    def merged_dict(self):
        return {**self.property.dict(), **self.additional_details.dict()}
